#ifndef STRUCTS_H
#define STRUCTS_H

/**********************************************
 *          the Beos Reversing Engine         * 
 *                  by Zadig                  *
 *                                            *
 * This file contains all Structures used     *
 * in the project.                            *
 **********************************************/

/*! \file structs.h
    \brief Contains all structures used in RevEng.
*/

/*enums*/
typedef enum
{
  OBJ_TYPE_UNKNOWN,        /*!< undefined object */
  OBJ_TYPE_STRING,         /*!< object is a string */
  OBJ_TYPE_OBJECT,         /*!< object is a variable, param... */
  OBJ_TYPE_MESSAGE,        /*!< object is a message */
  OBJ_TYPE_PARAM,          /*!< object is a parameter */
  OBJ_TYPE_CLASS,          /*!< object is a class */
  OBJ_TYPE_FUNCTION,       /*!< object is a function */
  OBJ_TYPE_INSTRUCTION,    /*!< object is an instruction */
  OBJ_TYPE_REFERENCE,      /*!< object is a reference */
  OBJ_TYPE_VAR_REFERENCE,  /*!< object is a reference to a variable or param */
  OBJ_TYPE_MAX
} RE_ObjectType;

/* possible result types for an asm instruction */
typedef enum
{
  RESULT_TYPE_INVALID,        /*!< invalid result */
  RESULT_TYPE_DIRECT_VALUE,   /*!< value that is put in second param (immediate addressing)*/
  RESULT_TYPE_INDIRECT_VALUE, /*!< value that is put in first param (direct or indirect addressing) */
  RESULT_TYPE_LOCAL,          /*!< value refers to a local variable */
  RESULT_TYPE_GLOBAL,         /*!< value refers to a global variable (not implemented) */
  RESULT_TYPE_PARAM,          /*!< value refers to a parameter */
  RESULT_TYPE_MESSAGE,        /*!< value is a message */
  RESULT_TYPE_JUMP,           /*!< value is a jump address */
  RESULT_TYPE_MAX
} RE_ResultType;

typedef enum
{
  REG_EAX,   /*!< Register eax */
  REG_EBX,   /*!< Register ebx */
  REG_ECX,   /*!< Register ecx */
  REG_EDX,   /*!< Register edx */
  REG_EBP,   /*!< Register ebp */
  REG_ESP,   /*!< Register esp */
  REG_EDI,   /*!< Register edi */
  REG_ESI,   /*!< Register esi */
  REG_MAX   
} RE_Registers;

typedef enum
{
  DIRECT_REF,  /*!< Direct reference */
  INDIRECT_REF /*!< Indirect reference */
} RE_RefType;

typedef enum
{
  MAP_STRINGS, /*!< Strings mapping index */
  MAP_SYMBOLS, /*!< Symbols mapping index */
  MAP_DYNSYM,  /*!< Dynamic symbols mapping index */
  MAP_REFS,    /*!< References mapping index */
  MAP_JUMPS,   /*!< Jumps mapping index */
  MAP_MESSAGES,/*!< Messages mapping index (BeOS specific) */
  MAP_PARAMS,  /*!< Params sizes index */
  MAP_MAX 
} RE_mappings;
/*end defines for structs*/

typedef struct _Reference Reference;   /*!< Reference struct (chainlist) */
struct _Reference
{
  uint32       addr; /*!< Address of reference */
  RE_RefType   type; /*!< Reference type */
  Reference* next;
} ;

typedef struct
{
  Reference *list_Ref;  /*!< Chainlist entry point */
} LIST_Reference;

/* chainlist struct used for all symbols */
typedef struct _Mapping Mapping; /*!< Mapping struct (chainlist) */
struct _Mapping
{
  uint32          addr;                   /*!< Addr of symbol or message value */
  uint8           label[STRING_MAX_SIZE]; /*!< Label of symbol */
  RE_ObjectType   type;                   /*!< Object type */
  LIST_Reference  references;             /*!< References to the object */
  uint8           isPrinted;              /*!< Symbol is already printed */
  Mapping         *next;                  /*!< Next object */
  Mapping         *prev;                  /*!< Previous object */
} ;

typedef struct
{
  Mapping *list_Map;    /*!< Chainlist entry point */
  Mapping *tail;        /*!< Tail of chainlist */
} LIST_Mapping;

typedef struct
{
  uint8  name[STRING_MAX_SIZE];
  uint32 addr;
  uint32 size;
} Section;

typedef struct
{ 
  uint8         name[REG_MAX_SIZE];
  uint32        value;
  RE_ResultType type;
  uint8         isValid;
} Register;

typedef struct
{
  Register registers[REG_MAX];
  uint32   addr;
  uint8    free;
} RegisterContextBase;

typedef struct
{
  RegisterContextBase context[MAX_REG_CNTX];
  uint32              nextCntx;
} RegisterContext;

typedef struct
{
  uint32 addr;
  uint32 value;
} PLT_Jump;

/* parameters structs */
typedef struct
{
  uint8 name[STRING_MAX_SIZE];
  uint8 offset;
} Parameter;

typedef struct
{
  uint8     isValid;
  uint8     functionIsC;   /* function is C or C++ */
  uint8     functionHasMessage;  
  uint8     nbParams;
  Parameter param[MAX_PARAMS];
} Function_Params;

typedef struct
{
   uint8    name[STRING_MAX_SIZE];
   uint32   size;
} Parameter_Size;

/* global struct */
typedef struct
{
  uint32  cep;
  uint8   exe_FileName[STRING_MAX_SIZE];
  uint8   dest_FileName[STRING_MAX_SIZE];
  uint8   asm_SrcFileName[STRING_MAX_SIZE];
  uint8   objdump_FileName[STRING_MAX_SIZE];
  Section sections[MAX_SECTIONS];
  uint8   splitOutput;
  uint8   ctags;
  uint8   printPlt;
} Settings ;


#endif
