/******************************************
 *      the Beos Reversing Engine         * 
 *              by Zadig                  *
 *                                        *
 * This file contains functions useful    *
 * to obtain "usable" files from objdump. *
 ******************************************/
/*! \file output.c
    \brief This is where output files are generated.
*/


/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "misc.h"
#include "output.h"
#include "chainlist.h"

/*****************************************************************************/
/*! \fn int32 RE_PrintHelp(void)
 *  \brief Printf RevEng help on stdout.
 *
 *****************************************************************************/
int32 RE_PrintHelp(void)
{
   printf("RevEng Error while reading parameters.\n");
   printf("\nusage:\n");
   printf("reveng [options] 'exe filename' 'dest filename'\n\n");
   printf("-split: Creates an 'exe filename' file for disassembly\n");    
   printf("        and an 'exe filename.sym' file for symbols.\n");
   printf("-plt:   Prints plt section (not usefull in most cases).\n");
   printf("-ctags: Generate tags to ease navigation in the disassembly\n");
   printf("        with vim.\n");
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_CreateOutputFile(Settings *projectSettings,
                                  LIST_Mapping *globalMapping)
 * \brief Creates the output file(s).
 *
 * This is the function that is called to create the output
 * files. It calls all other functions that print specific
 * parts of the target:
 * - If split option is enabled then creates 2 files.
 * - Prints RevEng banner.
 * - Prints all strings referenced in the target.
 * - Prints all classes prototypes of the target.
 * - Prints all functions names of the target.
 * - Finally prints the code of the target.
 *****************************************************************************/
int32 RE_CreateOutputFile(Settings *projectSettings,
                          LIST_Mapping *globalMapping)
{
   int32 i_ret;
   uint8 sz_SymbolsFile[STRING_MAX_SIZE];
   FILE *f_DisasmFile, *f_SymbolsFile;

   f_DisasmFile = fopen(projectSettings->dest_FileName,"w+");
   if(f_DisasmFile == NULL)
      return(FILE_ERROR_OPEN);

   if(projectSettings->splitOutput == Z_TRUE)
   {
      sprintf(sz_SymbolsFile,"%s.sym",projectSettings->dest_FileName);
      f_SymbolsFile = fopen(sz_SymbolsFile,"w+");
      if(f_SymbolsFile == NULL)
         return(FILE_ERROR_OPEN);
   }
   else
   {
      f_SymbolsFile = f_DisasmFile;
   }

   i_ret = RE_PrintHeader(f_DisasmFile);
   if(i_ret != Z_OK)
      return(i_ret);

   if(projectSettings->splitOutput == Z_TRUE)
   {
      i_ret = RE_PrintHeader(f_SymbolsFile);
      if(i_ret != Z_OK)
         return(i_ret);
   }

   /*i_ret = RE_PrintSections(f_OutputFile, projectSettings);
     if(i_ret != Z_OK)
     return(i_ret);
    */

   i_ret = RE_PrintStrings(f_SymbolsFile, projectSettings, 
                           &globalMapping[MAP_STRINGS]);
   if(i_ret != Z_OK)
      return(i_ret);

   i_ret = RE_PrintObjects(f_SymbolsFile, projectSettings,
                           &globalMapping[MAP_SYMBOLS]);
   if(i_ret != Z_OK)
      return(i_ret);

   i_ret = RE_PrintClasses(f_SymbolsFile, projectSettings, 
                           &globalMapping[MAP_SYMBOLS]);
   if(i_ret != Z_OK)
      return(i_ret);

   i_ret = RE_PrintFunctions(f_SymbolsFile, &globalMapping[MAP_SYMBOLS]);
   if(i_ret != Z_OK)
      return(i_ret);

   i_ret = RE_PrintCode(f_DisasmFile, projectSettings, globalMapping);
   if(i_ret != Z_OK)
      return(i_ret);

   fclose(f_DisasmFile);
   if(projectSettings->splitOutput == Z_TRUE)
      fclose(f_SymbolsFile);

   if(projectSettings->ctags)
   {
      i_ret = RE_CreateTagsFile(projectSettings->dest_FileName, 
                                &globalMapping[MAP_SYMBOLS]);
      if(i_ret != Z_OK)
         return(i_ret);
   }
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintHeader(FILE *f_OutputFile)
 *  \brief Prints RevEng's banner in a file.
 *
 *****************************************************************************/
int32 RE_PrintHeader(FILE *f_OutputFile)
{
   uint8 sz_Header[256];

   DBG_PRINT("Entering RE_PrintHeader...\n"); 

   RE_GetHeader(sz_Header);
   fprintf(f_OutputFile, "%s", sz_Header);  

   DBG_PRINT("End of RE_PrintHeader\n"); 
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintSections(FILE *f_OutputFile, Settings *projectSettings)
 *  \brief print sections list in a file.
 *
 *  This function is not yet implemented.
 *****************************************************************************/
int32 RE_PrintSections(FILE *f_OutputFile, Settings *projectSettings)
{
   DBG_PRINT("Entering RE_PrintSections...\n"); 
   fprintf(f_OutputFile,"\n[Elf Header]\n");  
   DBG_PRINT("End of RE_PrintSections\n");   

   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintCode(FILE *f_OutputFile, Settings *projectSettings, 
                           LIST_Mapping *globalMapping)
 *  \brief Prints the code of the target in a file 
 *
 *  This function prints the code of the target with sections
 *  names, functions names, objects and jumps references.
 *****************************************************************************/
int32 RE_PrintCode(FILE *f_OutputFile, Settings *projectSettings, 
                   LIST_Mapping *globalMapping)
{
   FILE *f_AsmFile;
   uint8 sz_CurLine[STRING_MAX_SIZE];
   uint8 b_End, b_ProjDone, b_JmpsDone, b_JumpPlt;
   int32 i, i_ret, i_NbRef;
   uint32 i_CurAddr;
   Mapping *curMap, *curMapRef, *curMapJmps;
   Reference *curRef;

   DBG_PRINT("Entering RE_PrintCode...\n");     
   f_AsmFile = fopen(projectSettings->asm_SrcFileName,"r");
   if(f_OutputFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }

   fprintf(f_OutputFile,"\n[Code Disassembly]\n");  

   curMap = globalMapping[MAP_SYMBOLS].list_Map;
   curMapJmps = globalMapping[MAP_JUMPS].list_Map;
   curMapRef = globalMapping[MAP_REFS].list_Map;
   b_JumpPlt = Z_FALSE;
   i_ret = (int32)fgets(sz_CurLine, STRING_MAX_SIZE, f_AsmFile);
   while(i_ret != 0)
   {  
      /* looks for definition and references */    
      sscanf(sz_CurLine,"%X",(unsigned int*) &i_CurAddr);

      /* prints if begining of section */
      i = 0;
      while(i < MAX_SECTIONS)
      {
         if(i_CurAddr == projectSettings->sections[i].addr)
         {
            if( (strcmp(projectSettings->sections[i].name, ".plt") == 0) &&
                  (projectSettings->printPlt == Z_FALSE) )
            {
               b_JumpPlt = Z_TRUE;
               break;
            }
            else
            {
               fprintf(f_OutputFile,"\nBegining of section %s...\n",
                       projectSettings->sections[i].name);
               b_JumpPlt = Z_FALSE;
               break;
            }
         }
         i++;
      }

      /* don't print plt section */    
      if(b_JumpPlt == Z_TRUE)
      {
         i_ret = (int32)fgets(sz_CurLine, STRING_MAX_SIZE, f_AsmFile);
         continue;
      }
      b_ProjDone = Z_FALSE;
      b_JmpsDone = Z_FALSE;
      b_End = Z_FALSE;

      while(b_End == Z_FALSE)
      {
         if( (curMap == NULL) ||
               (curMap->addr > i_CurAddr) )
         {
            b_ProjDone = Z_TRUE;
         }
         if( (curMapJmps == NULL) ||
               (curMapJmps->addr > i_CurAddr) )
         {
            b_JmpsDone = Z_TRUE;
         }
         if((b_JmpsDone == Z_TRUE) &&
               (b_ProjDone == Z_TRUE) )
         {
            b_End = Z_TRUE;
            continue;
         }

         if( (b_ProjDone == Z_FALSE) &&
               (curMap->addr == i_CurAddr))
         {
            if(curMap->type == OBJ_TYPE_FUNCTION)
               fprintf(f_OutputFile,"\nFunction %s\n",curMap->label);
            else if(curMap->type == OBJ_TYPE_CLASS)
               fprintf(f_OutputFile,"\nClass %s\n",curMap->label);
            else if(curMap->type == OBJ_TYPE_OBJECT)
               fprintf(f_OutputFile,"\nObject %s\n",curMap->label);

            /* prints where it is referenced */
            i_NbRef = 0;      
            curRef = curMap->references.list_Ref;
            while(curRef != NULL)
            {
               if(i_NbRef == 0)
               {
                  if(curMap->type == OBJ_TYPE_INSTRUCTION)
                  {
                     fprintf(f_OutputFile,
                        "\nReferenced by (conditionnal) jump(s) at Address(es):\n");
                  }
                  else
                  {
                     fprintf(f_OutputFile,"\nReferenced at Address(es):\n");
                  }
                  fprintf(f_OutputFile,"\t%.8X  ",(unsigned int)curRef->addr);
               }
               else
               {
                  fprintf(f_OutputFile,"%.8X  ",(unsigned int)curRef->addr);
               }

               i_NbRef++;
               if( ((i_NbRef % 5) == 0) && (i_NbRef != 0) )
               {
                  fprintf(f_OutputFile,"\n\t");          
               }
               curRef = curRef->next;          
            }
            if(i_NbRef != 0)
            {
               fprintf(f_OutputFile,"\n");
            }

         } /* end if(curMap->addr == i_CurAddr) */        

         if( (b_JmpsDone == Z_FALSE) &&
               (curMapJmps->addr == i_CurAddr))
         {
            /* prints where it is referenced */
            i_NbRef = 0;      
            curRef = curMapJmps->references.list_Ref;
            while(curRef != NULL)
            {
               if(i_NbRef == 0)
               {
                  fprintf(f_OutputFile,
                     "\nReferenced by (conditionnal) jump(s) at Address(es):\n");
                  fprintf(f_OutputFile,"\t%.8X  ",(unsigned int)curRef->addr);
               }
               else
               {
                  fprintf(f_OutputFile,"%.8X  ",(unsigned int)curRef->addr);
               }

               i_NbRef++;
               if( ((i_NbRef % 5) == 0) && (i_NbRef != 0) )
               {
                  fprintf(f_OutputFile,"\n\t");          
               }
               curRef = curRef->next;          
            }
            if(i_NbRef != 0)
            {
               fprintf(f_OutputFile,"\n");
            }
         } /* end if(curMapJmps->addr == i_CurAddr) */        

         if(b_ProjDone == Z_FALSE)
         {    
            curMap = curMap->next;        
         }
         if(b_JmpsDone == Z_FALSE) 
         {    
            curMapJmps = curMapJmps->next;        
         }
      } /* end while(curMap) */

      while( (curMapRef != NULL) && 
            (curMapRef->addr <= i_CurAddr) )
      {
         if(curMapRef->addr == i_CurAddr) 
         {
            if( (curMapRef->type != OBJ_TYPE_VAR_REFERENCE) &&
                (curMapRef->type != OBJ_TYPE_MESSAGE) )
            {
               fprintf(f_OutputFile,"%s\n",curMapRef->label);
            }
            else               
               break;

            curMapRef = curMapRef->next;
         }
      }

      sz_CurLine[strlen(sz_CurLine)-1] = '\0';
      fprintf(f_OutputFile,"%s",sz_CurLine);    
      if( (curMapRef != NULL) &&
            (curMapRef->addr == i_CurAddr) )
      {
         if( (curMapRef->type == OBJ_TYPE_VAR_REFERENCE) ||
               (curMapRef->type == OBJ_TYPE_MESSAGE) )
         {
            fprintf(f_OutputFile,"   %s\n",curMapRef->label);
            curMapRef = curMapRef->next;
         }
         else
            fputc((int)'\n',f_OutputFile);      
      }
      else
         fputc((int)'\n',f_OutputFile);

      i_ret = (int32)fgets(sz_CurLine, STRING_MAX_SIZE, f_AsmFile);
   }

   fclose(f_AsmFile);
   DBG_PRINT("End of RE_PrintCode\n");     
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintStrings(FILE *f_OutputFile, Settings *projectSettings,
                              LIST_Mapping *stringsMapping)
 *
 *  \brief prints strings list in a file.
 *
 *  This function prints all strings referenced in the target.
 *  Strings are printed in alphabetical order.
 *****************************************************************************/

int32 RE_PrintStrings(FILE *f_OutputFile, Settings *projectSettings,
                      LIST_Mapping *stringsMapping)
{
   int32 i_ret;
   Mapping *p_CurMap;

   DBG_PRINT("Entering RE_PrintStrings...\n");     
   p_CurMap = NULL;

   fprintf(f_OutputFile,"\n[Strings references]\n"); 
   i_ret = LIST_GetLowerStringMap(stringsMapping, &p_CurMap);
   while(i_ret == Z_OK) 
   {
      fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurMap->addr, p_CurMap->label);
      p_CurMap->isPrinted = Z_TRUE;
      i_ret = LIST_GetLowerStringMap(stringsMapping, &p_CurMap);
   }

   DBG_PRINT("End of RE_PrintStrings\n");      
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintObjects(FILE *f_OutputFile, Settings *projectSettings,
                              LIST_Mapping *stringsMapping)
 *
 *  \brief prints objects (globals variables) list in a file.
 *
 *  This function prints all objects defined in the target.
 *  Objects are printed in alphabetical order.
 *****************************************************************************/

int32 RE_PrintObjects(FILE *f_OutputFile, Settings *projectSettings,
                      LIST_Mapping *symbolsMapping)
{
   Mapping *p_CurMap;

   DBG_PRINT("Entering RE_PrintObjects...\n");     
   p_CurMap = symbolsMapping->list_Map;

   fprintf(f_OutputFile,"\n[Objects references]\n"); 
   while(p_CurMap != NULL) 
   {
      if( (p_CurMap->type == OBJ_TYPE_OBJECT) &&
          (strstr(p_CurMap->label, RTTI_SYMBOL) == NULL) )
         fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurMap->addr, 
                 p_CurMap->label);
      
      p_CurMap = p_CurMap->next;
   }

   DBG_PRINT("End of RE_PrintObjects\n");      
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintClasses(FILE *f_OutputFile, Settings *projectSettings,
                              LIST_Mapping *projectMapping)
 *
 * \brief Prints classes list in a file.
 *
 * This function prints all classes in alphabetical order, and
 * all methods of each class in alphabetical order.
 *
 *****************************************************************************/
int32 RE_PrintClasses(FILE *f_OutputFile, Settings *projectSettings,
                      LIST_Mapping *projectMapping)
{
   int32 i_ret;
   Mapping *p_CurClass, *p_CurMethod;
   uint8 sz_ClassName[STRING_MAX_SIZE];

   DBG_PRINT("Entering RE_PrintClasses...\n");     
   fprintf(f_OutputFile,"\n[Classes references]\n");

   i_ret = LIST_GetLowerClassMap(projectMapping, &p_CurClass, sz_ClassName);
   while(i_ret == Z_OK)
   {    
      fprintf(f_OutputFile,"\nClass %s\n", sz_ClassName);

      /* looks for all properties of this class */
      i_ret = LIST_GetLowerMethodMap(projectMapping, &p_CurMethod, sz_ClassName);
      while(i_ret == Z_OK)
      {
         fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurMethod->addr, 
                 p_CurMethod->label);
         p_CurMethod->isPrinted = Z_TRUE;
         i_ret = LIST_GetLowerMethodMap(projectMapping, &p_CurMethod, sz_ClassName);
      } 
      i_ret = LIST_GetLowerClassMap(projectMapping, &p_CurClass, sz_ClassName);
   }

   DBG_PRINT("End of RE_PrintClasses\n");       
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_PrintFunctions(FILE *f_OutputFile, LIST_Mapping *projectMapping)
 * \brief Prints functions of the target in a file.
 *
 * This function prints all functions name of the target in a file
 * in alphabetical order.
 *****************************************************************************/
int32 RE_PrintFunctions(FILE *f_OutputFile, LIST_Mapping *projectMapping)
{
   int32 i_ret;
   Mapping *p_CurFunction;
   uint8 *psz_FunctionName;

   DBG_PRINT("Entering RE_PrintFunctions...\n");     
   fprintf(f_OutputFile,"\n[Functions references]\n");

   p_CurFunction = NULL;
   i_ret = LIST_GetLowerStringMap(projectMapping, &p_CurFunction);  
   while(i_ret == Z_OK)
   {
      /* looks for new class */
      /* don't process templates */
      psz_FunctionName = strstr(p_CurFunction->label,"<"); 
      if( (psz_FunctionName == NULL) && /* templates doesn' contain '<' and '>' */
          (p_CurFunction->isPrinted == Z_FALSE) && /* don't process if printed */
          (p_CurFunction->type == OBJ_TYPE_FUNCTION) && /* process only functions */
          (strstr(p_CurFunction->label, RTTI_SYMBOL) == NULL) )
      {
         fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurFunction->addr,  
                 p_CurFunction->label);
      }

      p_CurFunction->isPrinted = Z_TRUE;
      i_ret = LIST_GetLowerStringMap(projectMapping, &p_CurFunction);
   }

   DBG_PRINT("End of RE_PrintClasses\n");       
   return(Z_OK);
}

/*****************************************************************************/
/*! \fn int32 RE_CreateTagsFile
 * \brief Creates the tags file.
 *
 * This function creates a tags file containing all functions and
 * classes of the target.
 * This will be used by VIM (and other editors) to step inside functions
 * easily.
 *****************************************************************************/
int32 RE_CreateTagsFile(uint8 *psz_DisasmFileName, 
                        LIST_Mapping *projectMapping)
{
   Mapping *p_CurMap;
   uint8 *psz_NameEnd, *psz_NameStart;
   uint8  sz_TagName[STRING_MAX_SIZE];
   FILE  *f_TagsFile;

   DBG_PRINT("Entering RE_CreateTagsFile...\n");     
   
   /* creates the file */
   f_TagsFile = fopen("tags","w+");
   if(f_TagsFile == NULL)
      return(FILE_ERROR_OPEN);

   /* prints header */
   fprintf(f_TagsFile, CTAG_HEADER, RE_VERSION);

   /* prints functions */
   p_CurMap = projectMapping->list_Map;
   while(p_CurMap != NULL)
   {
      /* looks for new class */
      /* don't process templates */
      if( (strstr(p_CurMap->label,"<") == NULL) && /* templates contain '<' and '>' */
          (strstr(p_CurMap->label,"global ") == NULL) &&
          (strstr(p_CurMap->label," function") == NULL) )
      {
         /* gets tag name */
         /* for classes, the tag name is the method name */
         /* without params                               */
         psz_NameStart = strstr(p_CurMap->label, "::");
         if(psz_NameStart != NULL)
            psz_NameStart += 2;
         else
            psz_NameStart = p_CurMap->label;
         
         psz_NameEnd = strstr(psz_NameStart, "(");
         if(psz_NameEnd == NULL)                    /* no params */
         {
            if(psz_NameStart != p_CurMap->label)    /* class is not possible */
            {
               p_CurMap = p_CurMap->next;
               continue;
            }
            else
            {
               psz_NameEnd = p_CurMap->label + strlen(p_CurMap->label);
            }
         }
         
         strncpy(sz_TagName, psz_NameStart, 
               psz_NameEnd - psz_NameStart);
         sz_TagName[psz_NameEnd - psz_NameStart] = '\0';
         
         /* creates tag entry in file */
         if(p_CurMap->type == OBJ_TYPE_FUNCTION) 
         {
            fprintf(f_TagsFile,"%s\t%s\t/^Function %s$/\n", sz_TagName,  
                    psz_DisasmFileName, p_CurMap->label);
         }
         else if(p_CurMap->type == OBJ_TYPE_CLASS)
         {
            fprintf(f_TagsFile,"%s\t%s\t/^Class %s$/\n", sz_TagName,
                    psz_DisasmFileName, p_CurMap->label);
         }
      }
      p_CurMap = p_CurMap->next;
   }

   fclose(f_TagsFile);
   DBG_PRINT("End of RE_CreateTagsFile\n");       
   return(Z_OK);
}
